#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import serializers, errors, utils
from ..compat import six, long_type
from .core import LazyLoad, cache, Iterable
from .cache import cache_parent
from .volumes import Volume


class FSVolumeObject(LazyLoad):
    __slots__ = '_volume_fs_tunnel',
    _type_indicator = '_isdir'
    _cache_name_arg = 'path'

    class CreateRequestXML(serializers.XMLSerializableModel):
        _root = 'Item'

        type = serializers.XMLNodeField('Type')
        path = serializers.XMLNodeField('Path')

    class UpdateRequestXML(serializers.XMLSerializableModel):
        _root = 'Item'

        path = serializers.XMLNodeField('Path')
        replication = serializers.XMLNodeField('Replication')

    _project = serializers.XMLNodeField('Project')
    _volume = serializers.XMLNodeField('Volume')
    path = serializers.XMLNodeField('Path')
    _isdir = serializers.XMLNodeField('Isdir', type='bool')
    permission = serializers.XMLNodeField('permission')
    _replication = serializers.XMLNodeField('BlockReplications', parse_callback=int)
    length = serializers.XMLNodeField('Length', parse_callback=long_type)
    quota = serializers.XMLNodeField('Quota', parse_callback=long_type)
    block_size = serializers.XMLNodeField('BlockSize', parse_callback=long_type)
    owner = serializers.XMLNodeField('Owner')
    group = serializers.XMLNodeField('Group')
    creation_time = serializers.XMLNodeField('CreationTime', type='rfc822')
    access_time = serializers.XMLNodeField('AccessTime', type='rfc822')
    last_modified_time = serializers.XMLNodeField('ModificationTime', type='rfc822')
    symlink = serializers.XMLNodeField('Symlink')
    sign_url = serializers.XMLNodeField('URL')

    @classmethod
    def _get_base_cls(cls):
        return FSVolumeObject

    @classmethod
    def _get_dir_cls(cls):
        return FSVolumeDir

    @classmethod
    def _get_file_cls(cls):
        return FSVolumeFile

    @classmethod
    def _get_objects_cls(cls):
        return FSVolumeObjects

    @staticmethod
    def _filter_cache(_, **kwargs):
        isdir = kwargs.get('_isdir')
        return isdir is not None and isdir != 'UNKNOWN'

    @utils.experimental('Volume2 is still experimental. Usage in production environment is strongly opposed.')
    @cache
    def __new__(cls, *args, **kwargs):
        isdir = kwargs.get('_isdir')

        base_cls = cls._get_base_cls()
        if cls is not base_cls and issubclass(cls, base_cls):
            return object.__new__(cls)
        if isdir is not None:
            if isdir == 'UNKNOWN':
                return object.__new__(base_cls)
            return object.__new__(cls._get_dir_cls() if isdir else cls._get_file_cls())

        obj = base_cls(_isdir='UNKNOWN', **kwargs)
        obj.reload()
        return base_cls(**obj.extract())

    def _name(self):
        return self.path

    def _set_state(self, name, parent, client):
        self.__init__(path=name, _parent=parent, _client=client)

    def split(self):
        return self.path.rsplit('/', 1)

    @property
    def basename(self):
        _, fn = self.split()
        return fn

    @property
    def dirname(self):
        dn, _ = self.split()
        return dn

    @property
    def volume(self):
        return self.parent

    @property
    def is_root(self):
        return self.path == '/' + self.parent.name

    def reload(self):
        # check if the volume path is the root
        if self.is_root:
            return

        schema_name = self.parent._get_schema_name()
        params = {}
        if schema_name is not None:
            params["curr_schema"] = schema_name

        resp = self._client.get(
            self.parent.resource(),
            action='meta',
            params=params,
            headers={'x-odps-volume-fs-path': self.path},
        )
        self.parse(self._client, resp, obj=self)

        self._loaded = True

    @staticmethod
    def _normpath(path):
        path = path.rstrip('/')
        i = 0
        parts = []
        start = 0
        while i < len(path):
            if path[i] == '/' or i == len(path) - 1:
                chunk = path[start:i + 1]
                start = i + 1
                if chunk in ['', '/', '.', './']:
                    # do nothing
                    pass
                elif chunk in ['..', '../']:
                    if len(parts):
                        parts = parts[:len(parts) - 1]
                    else:
                        parts.append(chunk)
                else:
                    parts.append(chunk)
            i += 1
        if path.startswith('/'):
            return '/' + ''.join(parts)
        return ''.join(parts)

    def _del_cache(self, path):
        root_objs = self._get_objects_cls()(parent=self.volume.root, client=self._client)
        if not path.startswith('/'):
            path = self.path + '/' + path.lstrip('/')
        del root_objs[path]

    def move(self, new_path, replication=None):
        """
        Move current path to a new location.

        :param new_path: target location of current file / directory
        :param replication: number of replication
        """
        if not new_path.startswith('/'):
            new_path = self._normpath(self.dirname + '/' + new_path)
        else:
            new_path = self._normpath(new_path)
        if new_path == self.path:
            raise ValueError('New path should be different from the original one.')
        update_def = self.UpdateRequestXML(path=new_path)
        if replication:
            update_def.replication = replication
        headers = {
            'Content-Type': 'application/xml',
            'x-odps-volume-fs-path': self.path,
        }

        schema_name = self.parent._get_schema_name()
        params = {}
        if schema_name is not None:
            params["curr_schema"] = schema_name

        self._client.put(
            self.parent.resource(),
            action='meta',
            params=params,
            headers=headers,
            data=update_def.serialize(),
        )

        self._del_cache(self.path)
        self.path = new_path
        self.reload()

    def _create_volume_fs_tunnel(self, endpoint=None, quota_name=None):
        if self._volume_fs_tunnel is not None:
            return self._volume_fs_tunnel

        from ..tunnel import VolumeFSTunnel

        self._volume_fs_tunnel = VolumeFSTunnel(
            client=self._client,
            project=self.project,
            endpoint=endpoint or self.project._tunnel_endpoint,
            quota_name=quota_name,
        )
        return self._volume_fs_tunnel


@cache_parent
class FSVolumeDir(FSVolumeObject):
    """
    VolumeFSDir represents a directory under a file system volume in ODPS.
    You can use ``create_dir`` to create a sub-directory, ``open_reader`` to open a file to read,
    ``open_writer`` to write a file and ``delete`` to remove. Following operations are also supported.

    >>> # iterate over all files and sub-directories
    >>> for o in fs_dir:
    >>>     print(o.path)
    >>> # check if a file exists in current volume
    >>> print(file_name in fs_dir)
    >>> # get a file/directory object
    >>> file_obj = fs_dir[file_name]
    """
    def __init__(self, **kw):
        super(FSVolumeDir, self).__init__(**kw)
        self._isdir = True

    @property
    def objects(self):
        return self._get_objects_cls()(parent=self, client=self._client)

    def create_dir(self, path):
        """
        Creates and returns a sub-directory under the current directory.
        :param str path: directory name to be created
        :return: directory object
        :rtype: :class:`odps.models.FSVolumeDir`
        """
        path = self.path + '/' + path.lstrip('/')
        dir_def = self.CreateRequestXML(type='directory', path=path)
        headers = {'Content-Type': 'application/xml'}
        self._client.post(
            self.parent.resource(),
            headers=headers,
            data=dir_def.serialize(),
            curr_schema=self.parent._get_schema_name(),
        )

        dir_object = type(self)(path=path, parent=self.parent, client=self._client)
        dir_object.reload()
        return dir_object

    def __contains__(self, item):
        return item in self.objects

    def __iter__(self):
        return self.objects.iterate()

    def __getitem__(self, item):
        return self.objects[item]

    def delete(self, recursive=False):
        """
        Delete current directory.

        :param recursive: indicate whether a recursive deletion should be performed.
        """
        params = {'recursive': recursive}
        headers = {'x-odps-volume-fs-path': self.path}
        self._del_cache(self.path)
        self._client.delete(
            self.parent.resource(),
            params=params,
            headers=headers,
            curr_schema=self.parent._get_schema_name(),
        )

    def open_reader(self, path, **kw):
        """
        Open a volume file and read contents in it.

        :param str path: file name to be opened
        :param start: start position
        :param length: length limit
        :param quota_name: name of tunnel quota
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :return: file reader

        :Example:
        >>> with fs_dir.open_reader('file') as reader:
        >>>     [print(line) for line in reader]
        """
        endpoint = kw.pop('endpoint', None)
        quota_name = kw.pop('quota_name', None)
        tunnel = self._create_volume_fs_tunnel(endpoint=endpoint, quota_name=quota_name)
        path = self.path.lstrip('/')[len(self.parent.name):].lstrip('/') + '/' + path.lstrip('/')
        return tunnel.open_reader(self.parent, path, **kw)

    def open_writer(self, path, replication=None, **kw):
        """
        Open a volume file and write contents into it.

        :param str path: file name to be opened
        :param quota_name: name of tunnel quota
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :return: file reader

        :Example:
        >>> with fs_dir.open_writer('file') as reader:
        >>>     writer.write('some content')
        """
        endpoint = kw.pop('endpoint', None)
        quota_name = kw.pop('quota_name', None)
        tunnel = self._create_volume_fs_tunnel(endpoint=endpoint, quota_name=quota_name)
        vol_path = self.path.lstrip('/')[len(self.parent.name):].lstrip('/') + '/' + path.lstrip('/')
        return tunnel.open_writer(self.parent, vol_path, replication=replication, **kw)


@cache_parent
class FSVolumeFile(FSVolumeObject):
    def __init__(self, **kw):
        super(FSVolumeFile, self).__init__(**kw)
        self._isdir = False

    @property
    def replication(self):
        """
        Get / set replication number of the file.
        """
        return self._replication

    @replication.setter
    def replication(self, value):
        update_def = self.UpdateRequestXML(replication=value)
        headers = {
            'Content-Type': 'application/xml',
            'x-odps-volume-fs-path': self.path,
        }

        schema_name = self.parent._get_schema_name()
        params = {}
        if schema_name is not None:
            params["curr_schema"] = schema_name

        self._client.put(
            self.parent.resource(),
            action='meta',
            params=params,
            headers=headers,
            data=update_def.serialize(),
        )
        self.reload()

    def delete(self, **_):
        """
        Delete current file.
        """
        params = {'recursive': False}
        headers = {'x-odps-volume-fs-path': self.path}

        schema_name = self.parent._get_schema_name()
        if schema_name is not None:
            params["curr_schema"] = schema_name

        self._del_cache(self.path)
        self._client.delete(self.parent.resource(), params=params, headers=headers)

    def open_reader(self, **kw):
        """
        Open current file and read contents in it.

        :param start: start position
        :param length: length limit
        :param quota_name: name of tunnel quota
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :return: file reader

        :Example:
        >>> with fs_file.open_reader('file') as reader:
        >>>     [print(line) for line in reader]
        """
        endpoint = kw.pop('endpoint', None)
        quota_name = kw.pop('quota_name', None)
        tunnel = self._create_volume_fs_tunnel(endpoint=endpoint, quota_name=quota_name)
        path = self.path.lstrip('/')[len(self.parent.name):].lstrip('/')
        return tunnel.open_reader(self.parent, path, **kw)

    def open_writer(self, replication=None, **kw):
        endpoint = kw.pop('endpoint', None)
        quota_name = kw.pop('quota_name', None)
        tunnel = self._create_volume_fs_tunnel(endpoint=endpoint, quota_name=quota_name)
        path = self.path.lstrip('/')[len(self.parent.name):].lstrip('/')
        return tunnel.open_writer(self.parent, path, replication=replication, **kw)


class FSVolumeObjects(Iterable):
    marker = serializers.XMLNodeField('Marker')
    max_items = serializers.XMLNodeField('MaxItems', parse_callback=int)
    objects = serializers.XMLNodesReferencesField(FSVolumeObject, 'Item')

    @property
    def project(self):
        return self.parent.parent.parent

    @property
    def volume(self):
        return self.parent.parent

    @classmethod
    def _get_single_object_cls(cls):
        return FSVolumeObject

    def _get(self, name):
        path = name
        if not path.startswith(self.parent.path):
            path = self.parent.path + '/' + name.lstrip('/')
        return self._get_single_object_cls()(client=self._client, parent=self.volume, path=path)

    def __contains__(self, item):
        if isinstance(item, six.string_types):
            try:
                # as reload() will be done in constructor of VolumeFSObject, we return directly.
                return self._get(item)
            except errors.NoSuchObject:
                return False
        elif isinstance(item, self._get_single_object_cls()):
            obj = item
            try:
                obj.reload()
                return True
            except errors.NoSuchObject:
                return False
        else:
            return False

    def __iter__(self):
        return self.iterate()

    def iterate(self):
        params = {'expectmarker': 'true'}
        headers = {'x-odps-volume-fs-path': self.parent.path}

        schema_name = self.volume._get_schema_name()
        if schema_name is not None:
            params["curr_schema"] = schema_name

        def _it():
            last_marker = params.get('marker')
            if 'marker' in params and \
                    (last_marker is None or len(last_marker) == 0):
                return

            url = self.volume.resource()
            resp = self._client.get(url, params=params, headers=headers)

            r = type(self).parse(self._client, resp, obj=self, parent=self.volume)
            params['marker'] = r.marker

            return r.objects

        while True:
            objects = _it()
            if objects is None:
                break
            for obj in objects:
                yield obj


@cache_parent
class FSVolume(Volume):
    """
    FSVolume represents the new-fashioned file system volume in ODPS.
    You can use ``create_dir`` to create a directory, ``open_reader`` to open a file to read
    and ``open_writer`` to write a file. Following operations are also supported.

    >>> # iterate over all files and directories
    >>> for o in fs_volume:
    >>>     print(o.path)
    >>> # check if a file exists in current volume
    >>> print(file_name in fs_volume)
    >>> # get a file/directory object
    >>> file_obj = fs_volume[file_name]
    """
    __slots__ = '_root_dir',

    _dir_cls = FSVolumeDir

    def create_dir(self, path):
        """
        Creates and returns a directory under the current volume.
        :param str path: directory name to be created
        :return: directory object
        :rtype: :class:`odps.models.FSVolumeDir`
        """
        return self.root.create_dir(path)

    def __contains__(self, item):
        return item in self.root

    def __iter__(self):
        for it in self.root:
            yield it

    def __getitem__(self, item):
        if not item:
            return self.root
        return self.root[item]

    @property
    def path(self):
        return '/' + self.name

    @property
    def location(self):
        if self.type != Volume.Type.EXTERNAL:
            raise AttributeError
        return self.properties.get(Volume.EXTERNAL_VOLUME_LOCATION_KEY)

    @property
    def rolearn(self):
        if self.type != Volume.Type.EXTERNAL:
            raise AttributeError
        return self.properties.get(Volume.EXTERNAL_VOLUME_ROLEARN_KEY)

    def open_reader(self, path, **kw):
        """
        Open a volume file and read contents in it.

        :param str path: file name to be opened
        :param start: start position
        :param length: length limit
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :return: file reader

        :Example:
        >>> with volume.open_reader('file') as reader:
        >>>     [print(line) for line in reader]
        """
        return self.root.open_reader(path, **kw)

    def open_writer(self, path, replication=None, **kw):
        """
        Open a volume file and write contents into it.

        :param str path: file name to be opened
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :return: file reader

        :Example:
        >>> with volume.open_writer('file') as reader:
        >>>     writer.write('some content')
        """
        return self.root.open_writer(path, replication=replication, **kw)

    def delete(self, path, recursive=False):
        try:
            return self[path].delete(recursive=recursive)
        except TypeError:
            return self[path].delete()

    @property
    def root(self):
        if not self._root_dir:
            self._root_dir = self._dir_cls(path='/' + self.name, parent=self, client=self._client)
        return self._root_dir


# keep renames compatible
VolumeFSDir = FSVolumeDir
VolumeFSFile = FSVolumeFile



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""UDF runners implementing the local run framework.
"""

import sys
import csv
from datetime import datetime

from ... import udf
from ... import distcache
from ...compat import six
from . import utils


__all__ = ['get_default_runner']


def get_default_runner(udf_class, input_col_delim=',', null_indicator='NULL', stdin=None):
    """Create a default runner with specified udf class.
    """
    proto = udf.get_annotation(udf_class)
    in_types, out_types = parse_proto(proto)
    stdin = stdin or sys.stdin
    arg_parser = ArgParser(in_types, stdin, input_col_delim, null_indicator)
    stdin_feed = make_feed(arg_parser)
    collector = StdoutCollector(out_types)
    ctor = _get_runner_class(udf_class)
    return ctor(udf_class, stdin_feed, collector)


def simple_run(udf_class, args):
    """
    """
    proto = udf.get_annotation(udf_class)
    in_types, out_types = parse_proto(proto)
    feed = direct_feed(args)
    collector = DirectCollector(out_types)
    ctor = _get_runner_class(udf_class)
    runner = ctor(udf_class, feed, collector)
    runner.run()
    return collector.results


def initialize():
    """Initialize the local run environment.
    """
    distcache.get_cache_table = utils.get_cache_table


def _get_types(types):
    entries = []
    for t in types.split(','):
        t = t.strip()
        if t not in _allowed_data_types:
            raise Exception('type not in '+ ','.join(_allowed_data_types))
        entries.append(_type_registry[t])
    return entries


def _get_in_types(types):
    if types == '':
        return []
    return _get_types(types) if types != '*' else [_type_registry[types], ]


def _get_runner_class(udf_class):
    if udf.BaseUDAF in udf_class.__mro__:
        ctor = UDAFRunner
    elif udf.BaseUDTF in udf_class.__mro__:
        ctor = UDTFRunner
    else:
        ctor = UDFRunner
    return ctor


def parse_proto(proto):
    tokens = proto.lower().split('->')
    if len(tokens) != 2:
        raise Exception('@annotate(' + proto + ') error')
    return _get_in_types(tokens[0].strip()), _get_types(tokens[1].strip())


def make_feed(arg_parser):
    for r in arg_parser.parse():
        yield r


def direct_feed(args):
    for a in args:
        yield a


class ArgParser(object):

    NULL_INDICATOR = 'NULL'

    def __init__(self, types, fileobj, delim=',', null_indicator='NULL'):
        self.types = types
        self.delim = delim
        self.null_indicator = null_indicator

        self.reader = csv.reader(fileobj, delimiter=delim)

    def parse(self):
        for record in self.reader:
            tokens = []
            for token in record:
                if token == self.null_indicator:
                    tokens.append(None)
                else:
                    tokens.append(token)
            if len(self.types) == 1 and self.types[0].typestr == '*':
                yield tokens
                continue
            if len(self.types) == 0 and len(tokens) == 0:
                yield ''
                continue

            if len(tokens) != len(self.types):
                raise Exception('Schema error: %r' % record)
            yield map(lambda tp, data: tp.converter(data), self.types, tokens)


class ArgFormater(object):

    DELIM = '\t'
    NULL_INDICATOR = 'NULL'

    def __init__(self, types):
        self.types = types

    def format(self, *args):
        ret = self.DELIM.join([str(a) for a in args])
        return ret


class BaseCollector(object):
    """Basic common logic of collector.
    """
    def __init__(self, schema):
        self.schema = schema

    def _validate_records(self, *args):
        if len(args) != len(self.schema):
            raise Exception('Schema error: ' + repr(args))
        for i, a in enumerate(args):
            if a is None:
                continue
            elif not isinstance(a, self.schema[i].type):
                raise Exception('Schema error: ' + repr(args))

    def collect(self, *args):
        self._validate_records(*args)
        self.handle_collect(*args)


class StdoutCollector(BaseCollector):
    """Collect the results to stdout.
    """

    def __init__(self, schema):
        super(StdoutCollector, self).__init__(schema)
        self.formater = ArgFormater(schema)

    def handle_collect(self, *args):
        print(self.formater.format(*args))


class DirectCollector(BaseCollector):
    """Collect results which can be fetched via self.results into memory.
    """

    def __init__(self, schema):
        super(DirectCollector, self).__init__(schema)
        self.results = []

    def handle_collect(self, *args):
        if len(self.schema) == 1:
            self.results.append(args[0])
        else:
            self.results.append(args)


class BaseRunner(object):

    def __init__(self, udf_class, feed, collector):
        self.udf_class = udf_class
        self.feed = feed
        self.collector = collector
        # check signature
        self.obj = udf_class()


class UDFRunner(BaseRunner):

    def run(self):
        obj = self.obj
        collector = self.collector
        for args in self.feed:
            r = obj.evaluate(*args)
            collector.collect(r)


class UDTFRunner(BaseRunner):

    def run(self):
        obj = self.obj
        collector = self.collector
        def local_forward(*r):
            collector.collect(*r)
        obj.forward = local_forward
        for args in self.feed:
            obj.process(*args)
        obj.close()


class UDAFRunner(BaseRunner):

    def run(self):
        obj = self.obj
        collector = self.collector
        buf0 = obj.new_buffer()
        buf1 = obj.new_buffer()
        turn = True
        for args in self.feed:
            if turn:
                buf = buf0
                turn = False
            else:
                buf = buf1
                turn = True
            obj.iterate(buf, *args)
        merge_buf = obj.new_buffer()
        obj.merge(merge_buf, buf0)
        obj.merge(merge_buf, buf1)
        collector.collect(obj.terminate(merge_buf))



###########################################
###         Static type registry        ###

def register_type(enum, typestr, tp):
    type_obj = TypeEntry(typestr, tp)
    globals()[enum] = type_obj
    _type_registry[typestr] = type_obj


_type_registry = {
}


def _gen_converter(typestr, tp):
    def f(v):
        if v == "NULL" or v is None:
            return None
        if typestr in ('bigint', 'datetime'):
            return int(v)
        elif typestr == 'string':
            return str(v)
        return tp(v)
    return f


class TypeEntry(object):

    def __init__(self, typestr, tp):
        self.typestr = typestr
        self.type = tp
        self.converter = _gen_converter(typestr, tp)

register_type('TP_BIGINT',   'bigint',   six.integer_types)
register_type('TP_STRING',   'string',   six.string_types)
if sys.version_info[0] == 2:
    register_type('TP_DATETIME', 'datetime', six.integer_types)
else:
    register_type('TP_DATETIME', 'datetime', datetime)
register_type('TP_DOUBLE',   'double',   float)
register_type('TP_BOOLEAN',  'boolean',  bool)
register_type('TP_STAR',     '*',        lambda x: x)

_allowed_data_types = [k for k in _type_registry.keys() if k != '*']

###########################################


initialize()




# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import logging

from ..config import options  ## don't remove
from ..compat import six, Enum  ## don't remove
from ..ui import reload_instance_status, fetch_instance_group

logger = logging.getLogger(__name__)
runners = dict()


class RunnerType(Enum):
    MOCK = 'MOCK'
    LAMBDA = 'LAMBDA'
    XFLOW = 'XFLOW'
    XLIB = 'XLIB'
    SQL = 'SQL'
    DF = 'DF'
    PS = 'PS'


def node_runner(engine):
    def _decorator(cls):
        runners[engine] = cls
        return cls

    return _decorator


def create_node_runner(engine, algo_name, params, metas, engine_kw, ui, **kw):
    runner_type = RunnerType(metas.get('engine', 'XFLOW').upper())
    runner = runners[runner_type]
    return runner(engine, algo_name, params, metas, engine_kw, ui, **kw)


class BaseNodeRunner(object):
    def __init__(self, engine, algo_name, params, metas, engine_kw, ui, progress_proportion=1,
                 priority=None, group=None, **kw):
        self._engine = engine
        self._algo_name = algo_name
        self._gen_params = params
        self._instances = []
        self._progress_group = group
        self._metas = metas
        self._engine_kw = engine_kw
        self._ui = ui
        self._progress_portion = progress_proportion
        self._last_cmd = None
        self._cases = kw.get('_cases', [])
        self._output_models_only = kw.get('_output_models_only')

    def execute(self):
        logger.info('Executing {0}...'.format(self._algo_name))
        self.before_exec()
        if options.ml.dry_run:
            self.mock_exec()
        else:
            self.actual_exec()

            while not all(inst.is_terminated(retry=True) for inst in self._instances):
                self.refresh_progress()
                time.sleep(1)
            [inst.wait_for_success() for inst in self._instances]
            self._ui.inc(self._progress_portion)
        self.after_exec()

    def before_exec(self):
        pass

    def actual_exec(self):
        raise NotImplementedError

    def mock_exec(self):
        if not hasattr(self, '_cases') or self._cases is None:
            return
        for case in self._cases:
            case(self, self._gen_params)

    def after_exec(self):
        pass

    def add_instance(self, inst, show_log_view=True):
        self._engine._instances.append(inst)
        self._instances.append(inst)
        if logger.getEffectiveLevel() <= logging.INFO:
            logger.info('Instance ID: ' + inst.id)
            if show_log_view:
                logger.info('  Log view: ' + inst.get_logview_address())
        self._ui.status('Executing', 'execution details')

    def refresh_progress(self):
        for inst in self._instances:
            reload_instance_status(self._engine._odps, self._progress_group, inst.id)
        self._ui.update_group()

    @classmethod
    def _format_value(cls, value):
        if value is None:
            return None
        elif isinstance(value, bool):
            if value:
                return 'true'
            else:
                return 'false'
        elif isinstance(value, (list, set)):
            if not value:
                return None
            return ','.join([cls._format_value(v) for v in value])
        else:
            return str(value)


@node_runner(RunnerType.XFLOW)
class XFlowNodeRunner(BaseNodeRunner):
    def __init__(self, *args, **kwargs):
        super(XFlowNodeRunner, self).__init__(*args, **kwargs)
        self._sub_instance_set = set()

    @staticmethod
    def _is_param_valid(val):
        if val is None:
            return False
        if isinstance(val, (list, set, tuple)) and len(val) == 0:
            return False
        return True

    def before_exec(self):
        super(XFlowNodeRunner, self).before_exec()
        if 'core_num' in self._engine_kw:
            self._gen_params['coreNum'] = self._engine_kw['core_num']
        if 'core_mem' in self._engine_kw:
            self._gen_params['memSizePerCore'] = self._engine_kw['core_mem']
        if 'lifecycle' in self._engine_kw and not options.ml.dry_run:
            self._gen_params['lifecycle'] = self._engine_kw['lifecycle']

    def actual_exec(self):
        pai_project = self._metas.get('xflowProjectName', options.ml.xflow_project)
        xflow_name = self._metas.get('xflowName', self._algo_name)

        params = dict([(k, self._format_value(v)) for k, v in six.iteritems(self._gen_params)
                       if self._is_param_valid(v)])

        if self._output_models_only:
            params.pop('lifecycle', None)

        param_args = ' '.join(['-D%s="%s"' % (k, v) for k, v in six.iteritems(params)
                               if self._is_param_valid(v)])
        self._last_cmd = 'PAI -name %s -project %s %s;' % (xflow_name, pai_project, param_args)
        logger.info('Command: ' + self._last_cmd)

        inst = self._engine._odps.run_xflow(xflow_name, pai_project, params)
        self.add_instance(inst, show_log_view=False)

    def refresh_progress(self):
        insts = self._instances
        if insts:
            inst = insts[0]
            group_json = fetch_instance_group(self._progress_group)
            group_json.logview = inst.get_logview_address()
        for xflow_inst in insts:
            for inst_name, sub_inst in six.iteritems(self._engine._odps.get_xflow_sub_instances(xflow_inst)):
                if sub_inst.id not in self._sub_instance_set:
                    self._sub_instance_set.add(sub_inst.id)
                    logger.info('Sub Instance: {0} ({1})'.format(inst_name, sub_inst.id))
                    logger.info('  Log view: ' + sub_inst.get_logview_address())
                reload_instance_status(self._engine._odps, self._progress_group, sub_inst.id)
        self._ui.update_group()


@node_runner(RunnerType.SQL)
class SQLNodeRunner(BaseNodeRunner):
    def actual_exec(self):
        query = self._gen_params['sql']
        self._last_cmd = query
        logger.info('Command: ' + self._last_cmd)

        inst = self._engine._odps.run_sql(query)
        self.add_instance(inst)


try:
    from ..internal.ml.runners import *
except ImportError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import json
import re
import sys
import time
import warnings

from .. import errors, readers, utils
from ..compat import six, enum
from ..lib.monotonic import monotonic
from ..models import tasks
from .instance import Instance, InstanceArrowReader, InstanceRecordReader


DEFAULT_TASK_NAME = "AnonymousSQLRTTask"
PUBLIC_SESSION_NAME = "public.default"


class FallbackMode(enum.Enum):
    OFFLINE = 0
    INTERACTIVE = 1


class FallbackPolicy:
    def __init__(
        self,
        policy=None,
        always=False,
        noresource=False,
        unsupported=False,
        timeout=False,
        upgrading=False,
        generic=False,
    ):
        policies = set()
        if policy:
            if isinstance(policy, (set, list, tuple)):
                policies = set(policy)
            else:
                policy = policy.lower().strip()
                if policy == "default":
                    policy = "unsupported,upgrading,noresource,timeout"
                elif policy == "all":
                    always = True
                policies = set(s.strip() for s in policy.split(","))
        self.always = always
        self.noresource = noresource or always or "noresource" in policies
        self.unsupported = unsupported or always or "unsupported" in policies
        self.timeout = timeout or always or "timeout" in policies
        self.upgrading = upgrading or always or "upgrading" in policies
        self.generic = generic or always or "generic" in policies

    def get_mode_from_exception(self, exc_value):
        err_msg = str(exc_value)
        if isinstance(exc_value, errors.SQARetryError):
            return FallbackMode.INTERACTIVE
        elif "OdpsJobCancelledException" in err_msg or "Job is cancelled" in err_msg:
            return None
        elif self.always:
            return FallbackMode.OFFLINE
        elif self.unsupported and isinstance(exc_value, errors.SQAUnsupportedFeature):
            return FallbackMode.OFFLINE
        elif self.upgrading and isinstance(exc_value, (errors.SQAServiceUnavailable, errors.SQAAccessDenied)):
            return FallbackMode.OFFLINE
        elif self.noresource and isinstance(exc_value, errors.SQAResourceNotEnough):
            return FallbackMode.OFFLINE
        elif self.timeout and (
            isinstance(exc_value, errors.SQAQueryTimedout)
            or "Wait for cache data timeout" in err_msg
            or "Get select desc from SQLRTTask timeout" in err_msg
        ):
            return FallbackMode.OFFLINE
        elif self.generic and isinstance(exc_value, errors.SQAGenericError):
            return FallbackMode.OFFLINE
        return None

    def __repr__(self):
        policies = [
            s for s in ["generic", "unsupported", "upgrading", "noresource", "timeout"]
            if getattr(self, s, None)
        ]
        return "<FallbackPolicy %s>" % ",".join(policies)


@enum.unique
class SessionTaskStatus(enum.Enum):
    """
    Possible statuses of tasks executing inside a session.
    """
    Running = 2
    Failed = 4
    Terminated = 5
    Cancelled = 6
    Unknown = -1


TASK_STATUS_VALUES = {
    2: SessionTaskStatus.Running,
    4: SessionTaskStatus.Failed,
    5: SessionTaskStatus.Terminated,
    6: SessionTaskStatus.Cancelled
}


def _task_status_value_to_enum(task_status):
    return TASK_STATUS_VALUES.get(task_status, SessionTaskStatus.Unknown)


def _get_session_failure_info(task_results):
    try:
        taskname, result_txt = list(task_results.items())[0]
    except BaseException:
        return ''
    return result_txt


class SessionInstance(Instance):
    """
    This represents the instance
    created right after you call 'attach_session' or 'create_session'.
    Further SQL tasks has to be created using this instance.
    """

    __slots__ = ('_project', '_task_name', '_session_name')

    def __init__(self, **kw):
        if 'session_task_name' not in kw or 'session_project' not in kw:
            raise errors.InvalidArgument("Creating InSessionInstance without enough information.")
        self._task_name = kw.pop("session_task_name", "")
        self._project = kw.pop("session_project", None)
        self._session_name = kw.pop("session_name", "")
        super(SessionInstance, self).__init__(**kw)

    def wait_for_startup(self, interval=1, timeout=-1, retry=True, max_interval=None):
        """
        Wait for the session to startup(status changed to RUNNING).

        :param interval: time interval to check (unit seconds)
        :param timeout: wait timeout (unit seconds), < 0 means no timeout
        :param retry: if failed to query session status, should we retry silently
        :raise: :class:`odps.errors.WaitTimeoutError` if wait timeout and session is not started.
        :return: None
        """
        start_time = monotonic()
        end_time = start_time + timeout
        while not self.is_running(retry):
            if timeout > 0:
                if monotonic() > end_time:
                    raise errors.WaitTimeoutError(
                        "Waited %.1f seconds, but session is not started."
                        % (monotonic() - start_time),
                        instance_id=self.id,
                    )
            try:
                time.sleep(interval)
                if max_interval:
                    interval = min(interval * 2, max_interval)
            except KeyboardInterrupt:
                return

    def _parse_result_session_name(self, result_str):
        if not self._session_name:
            session_name_search = re.search("Session name: (.*)$", result_str)
            if session_name_search:
                self._session_name = session_name_search.group(1)

    def run_sql(self, sql, hints=None, **kwargs):
        task = tasks.SQLTask(query=utils.to_text(sql), **kwargs)
        task.update_sql_settings(hints)
        return self._create_internal_instance(task=task)

    @staticmethod
    def _check_is_select(sql_statement):
        try:
            splited = utils.split_sql_by_semicolon(sql_statement)
        except Exception as ex:
            warnings.warn(
                "Cannot split sql statement %s: %s" % (sql_statement, str(ex)), RuntimeWarning
            )
            return False
        return splited[-1].lower().strip(' \t\r\n(').startswith("select")

    def _create_internal_instance(self, task=None):
        project_name = self._project.name
        is_select = self._check_is_select(task.query.strip())

        self.parent._fill_task_properties(task)
        rquery = task.query
        if not rquery.endswith(";"):
            rquery = rquery + ";"
        query_object = {
            "query": rquery,
            "settings": json.loads(task.properties['settings'])
        }
        query_json = json.dumps(query_object)

        resp_content = self.put_task_info(
            self._task_name, "query", query_json, check_location=True
        )

        created_subquery_id = -1
        try:
            query_result = json.loads(resp_content)
            query_status = query_result["status"]
            if query_status != "ok":
                raise errors.ODPSError(
                    'Failed to run subquery: [' + query_status + "]: " + query_result["result"]
                )
            query_subresult = json.loads(query_result["result"])
            created_subquery_id = query_subresult["queryId"]
            if created_subquery_id == -1:
                raise errors.parse_instance_error(query_subresult)
        except KeyError as ex:
            six.raise_from(
                errors.ODPSError(
                    "Invalid Response Format: %s\n Response JSON:%s\n" % (str(ex), resp_content.decode())
                ), None
            )
        instance = InSessionInstance(
            session_project_name=project_name, session_task_name=self._task_name,
            name=self.id, session_subquery_id=created_subquery_id,
            session_instance=self, parent=self.parent, session_is_select=is_select,
            client=self._client,
        )
        return instance

    def reload(self):
        resp_text = self.get_task_info(self._task_name, "status")
        session_status = SessionTaskStatus.Unknown
        try:
            poll_result = json.loads(resp_text)
            self._parse_result_session_name(poll_result["result"])
            session_status = _task_status_value_to_enum(poll_result["status"])
        except BaseException:
            error_string = _get_session_failure_info(self.get_task_results())
            if error_string:
                self._status = Instance.Status.TERMINATED
                six.raise_from(errors.parse_instance_error(error_string), None)
            else:
                # this is a task meta info update problem. Just retry.
                self._status = Instance.Status.SUSPENDED
                return
        if session_status == SessionTaskStatus.Running:
            self._status = Instance.Status.RUNNING
        elif session_status == SessionTaskStatus.Cancelled:
            error_string = _get_session_failure_info(self.get_task_results())
            self._status = Instance.Status.TERMINATED
            raise errors.parse_instance_error(error_string)
        elif session_status == SessionTaskStatus.Failed:
            error_string = _get_session_failure_info(self.get_task_results())
            self._status = Instance.Status.TERMINATED
            raise errors.parse_instance_error(error_string)
        elif poll_result["status"] == SessionTaskStatus.Terminated:
            self._status = Instance.Status.TERMINATED
            raise errors.ODPSError("Session terminated.")
        else:
            self._status = Instance.Status.SUSPENDED


class InSessionTunnelReaderMixin(object):
    @property
    def schema(self):
        # is not available before open_reader().
        if self._download_session.schema is None:
            # open reader once to enforce schema fetched.
            tmprd = self._download_session.open_record_reader(0, 1)
            tmprd.close()
        return self._download_session.schema

    @property
    def count(self):
        # we can't count session results before it's
        # fully retrieved.
        return -1

    @property
    def status(self):
        # force reload to update download session status
        # this is for supporting the stream download of instance tunnel
        # without the following line will not trigger reload
        self._download_session.reload()
        return self._download_session.status

    def read(self, start=None, count=None, step=None,
             compress=False, columns=None):
        start = start or 0
        step = step or 1
        stop = None if count is None else start + step * count

        with self._download_session.open_record_reader(
                0, 1, compress=compress, columns=columns) as reader:
            for record in itertools.islice(reader, start, stop, step):
                yield record


class InSessionInstanceArrowReader(InSessionTunnelReaderMixin, InstanceArrowReader):
    pass


class InSessionInstanceRecordReader(InSessionTunnelReaderMixin, InstanceRecordReader):
    pass


class InSessionInstance(Instance):
    """
        This represents the instance created
        for SQL tasks that run inside a session. This instance is useful
        when fetching results.
    """

    __slots__ = (
        '_project_name', '_session_task_name', '_session', '_session_instance',
        '_is_select', '_subquery_id', '_report_result', '_report_warning',
        '_session_task_status',
    )

    def __init__(self, **kw):
        if (
            'session_task_name' not in kw
            or 'session_project_name' not in kw
            or 'session_instance' not in kw
            or 'session_subquery_id' not in kw
        ):
            raise errors.InvalidArgument("Creating InSessionInstance without enough information.")
        self._session_task_name = kw.pop("session_task_name", "")
        self._project_name = kw.pop("session_project_name", "")
        self._session_instance = kw.pop("session_instance", None)
        self._is_select = kw.pop("session_is_select", False)
        self._subquery_id = kw.pop("session_subquery_id", -1)
        self._report_result = ''
        self._report_warning = ''
        self._session_task_status = -1
        if self._subquery_id < 0:
            raise errors.InternalServerError("Subquery id not legal: " + str(self._subquery_id))
        super(InSessionInstance, self).__init__(**kw)

    @property
    def subquery_id(self):
        return self._subquery_id

    @utils.survey
    def _open_result_reader(self, schema=None, task_name=None, **kwargs):
        """
        Fetch result directly from odps. This way does not support limiting, and will cache
        all result in local memory. To achieve better performance and efficiency, use tunnel instead.
        """
        if not self._is_select:
            raise errors.InstanceTypeNotSupported("No results for non-select SQL.")
        self.reload()
        if not self.is_successful(retry=True):
            raise errors.ODPSError(
                'Cannot open reader, instance(%s) may fail or has not finished yet' % self.id)
        return readers.CsvRecordReader(schema, self._report_result)

    def _open_tunnel_reader(self, **kw):
        if not self._is_select:
            raise errors.InstanceTypeNotSupported(
                "InstanceTunnel cannot be opened at a non-select SQL Task."
            )

        while (self._subquery_id == -1) and (self._status != Instance.Status.TERMINATED):
            self.reload()

        if self._subquery_id == -1:
            raise errors.InternalServerError("SubQueryId not returned by the server.")

        kw.pop('reopen', False)
        arrow = kw.pop("arrow", False)
        endpoint = kw.pop('endpoint', None)
        kw['sessional'] = True
        kw['session_subquery_id'] = self._subquery_id
        if 'session_task_name' not in kw:
            kw['session_task_name'] = self._session_task_name

        tunnel = self._create_instance_tunnel(endpoint=endpoint)

        try:
            download_session = tunnel.create_download_session(instance=self, **kw)
        except errors.InternalServerError:
            e, tb = sys.exc_info()[1:]
            e.__class__ = Instance.DownloadSessionCreationError
            six.reraise(Instance.DownloadSessionCreationError, e, tb)

        self._download_id = download_session.id

        if arrow:
            return InSessionInstanceArrowReader(self, download_session)
        else:
            return InSessionInstanceRecordReader(self, download_session)

    def reload(self):
        resp_text = self.get_task_info(self._session_task_name, "result_" + str(self._subquery_id))
        try:
            query_result = json.loads(resp_text)
            query_status = query_result["status"]
            self._report_result = query_result["result"]
            self._report_warning = query_result["warnings"]
            self._session_task_status = _task_status_value_to_enum(query_status)
            if self._session_task_status in (
                SessionTaskStatus.Terminated, SessionTaskStatus.Failed, SessionTaskStatus.Cancelled
            ):
                self._status = Instance.Status.TERMINATED
            elif self._session_task_status == SessionTaskStatus.Running:
                self._status = Instance.Status.RUNNING
            else:
                self._status = Instance.Status.SUSPENDED
            self._subquery_id = int(query_result.get("subQueryId", -1))
        except BaseException as ex:
            raise errors.ODPSError(
                "Invalid Response Format: %s\n Response JSON:%s\n" % (str(ex), resp_text)
            )

    def is_successful(self, retry=False):
        """
        If the instance runs successfully.

        :return: True if successful else False
        :rtype: bool
        """

        if not self.is_terminated(retry=retry):
            return False
        if self._session_task_status in (SessionTaskStatus.Failed, SessionTaskStatus.Cancelled):
            return False
        return True

    def wait_for_success(self, interval=1, timeout=None, max_interval=None):
        """
        Wait for instance to complete, and check if the instance is successful.

        :param interval: time interval to check
        :param max_interval: if specified, next check interval will be
            multiplied by 2 till max_interval is reached.
        :param timeout: time
        :return: None
        :raise: :class:`odps.errors.ODPSError` if the instance failed
        """

        self.wait_for_completion(interval=interval, max_interval=max_interval, timeout=timeout)

        if not self.is_successful(retry=True):
            raise errors.parse_instance_error(self._report_result)

    def get_warnings(self):
        """
        Get the warnings reported by ODPS.

        :return: warning string if ever reported, or empty string for no warning.
        """

        self.reload()
        return self._report_warning

    def get_printable_result(self):
        """
        Get the result string that can be directly printed to screen.
        This should only be used for interactive display. The returning format is not guaranteed.

        :return: The printable result. On not completed or no result returned, will return empty string.
        :raise: :class:`odps.errors.ODPSError` if the instance failed.
        """

        self.reload()
        if self.is_terminated() and not self.is_successful():
            raise errors.parse_instance_error(self._report_result)
        return self._report_result



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def linear_svm_set_cost(algo, param_name, param_value):
    algo._parameters['positiveCost'].value = param_value
    algo._parameters['negativeCost'].value = param_value


"""
Random forests exporters
"""


def random_forest_algo_types(expr):
    algo_type = expr._params["algorithmType"]
    tree_num = int(expr._params["treeNum"])
    tree_num = tree_num - 1 if tree_num != 0 else 0
    if algo_type == "mix":
        return None
    elif algo_type == "id3":
        return '%d,%d' % (tree_num, tree_num)
    elif algo_type == "c45":
        return "0,0"
    elif "cart":
        return "0,%d" % tree_num
    return None


def random_forest_min_num_per(expr):
    ret = '-1'
    if expr._params['minNumPerInt'] is None:
        return ret

    min_num_per_int = expr._params['minNumPerInt']
    if min_num_per_int.strip() != '':
        if int(min_num_per_int) in (-1, 0):
            return ret
        else:
            return int(min_num_per_int) / 100.0
    return str(ret)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mmap
import os
import tempfile
from collections import OrderedDict
from urllib.parse import urlparse

import numpy as np
from mars.lib.filesystem import FileSystem

from ...core import ODPS
from ...compat import futures, BytesIO
from ...errors import NoSuchObject
from ...tunnel.volumetunnel import VolumeTunnel
from ...utils import to_binary, to_str


DIR_ERROR_MSG = (
    "Argument `path` illegal, expect to be "
    "odps:///**project**/volumes/**volume**/, "
    "got {}"
)

FILE_ERROR_MSG = (
    "Argument `path` illegal, expect to be "
    "odps:///**project**/volumes/**volume**/**file**, "
    "got {}"
)

DEFAULT_CHUNK_SIZE = 512 * 1024**2

META_FILE_NAME = "_meta_"


class _VolumeFileObject(object):
    def __init__(self, volume_parted, mode, temp_dir=None, chunk_size=None):
        self._volume_parted = volume_parted
        self._mode = mode
        self._temp_dir = temp_dir
        self._chunk_size = chunk_size

        # writer
        self._index = 0
        self._writer = None
        self._write_buffer = None
        self._meta = dict()

        # reader
        self._reader = None
        self._read_filename = None

    def _read_multi_volume_files(self):
        # get meta file first
        with self._volume_parted.open_reader(META_FILE_NAME) as meta_reader:
            content = to_str(meta_reader.read())
            meta = OrderedDict()
            length = 0
            for line in content.split("\n"):
                if "," in line:
                    filename, size = line.split(",", 1)
                    size = int(size)
                    meta[filename] = size
                    length += size

        temp_dir = self._temp_dir or tempfile.mkdtemp()
        write_filename = os.path.join(
            temp_dir, self._volume_parted.volume.name, self._volume_parted.name
        )
        os.makedirs(os.path.dirname(write_filename), exist_ok=True)

        with open(write_filename, "wb") as f:
            for size in meta.values():
                f.write(b"\0" * size)

        with open(write_filename, "rb+") as f:
            m = mmap.mmap(f.fileno(), length)
            offsets = [0] + np.cumsum(list(meta.values())).tolist()

            volume_part = self._volume_parted
            volume_tunnel = VolumeTunnel(
                client=volume_part._client,
                project=volume_part.project,
                endpoint=volume_part.project._tunnel_endpoint,
            )

            def _write(filename, start, end):
                with volume_tunnel.create_download_session(
                    volume_part.volume.name, volume_part.name, filename
                ).open() as reader:
                    content_part = reader.read()
                    m[start:end] = content_part

            executor = futures.ThreadPoolExecutor(8)
            fs = []
            for i, filename in enumerate(meta):
                future = executor.submit(_write, filename, offsets[i], offsets[i + 1])
                fs.append(future)
            [f.result() for f in fs]

        self._read_filename = write_filename
        self._reader = open(write_filename, "rb")

    def read(self, size=None):
        self._volume_parted.reload()
        if self._volume_parted.file_number == 1:
            if self._reader is None:
                self._reader = self._volume_parted.open_reader(
                    list(self._volume_parted.files)[0].name
                )
            return self._reader.read(size=size)
        else:
            if self._reader is None:
                self._read_multi_volume_files()
            return self._reader.read(size)

    def _ensure_writer_open(self):
        if self._writer is None:
            self._writer = self._volume_parted.open_writer()
            self._write_buffer = BytesIO()

    def _flush_to_volume_file(self):
        # write buffer full, write with a file
        filename = str(self._index)
        buffer = self._write_buffer
        content_part = buffer.getvalue()
        self._writer.write(filename, content_part)
        self._meta[filename] = buffer.tell()
        self._index += 1
        # clear write buffer
        self._write_buffer = BytesIO()

    def write(self, content):
        self._ensure_writer_open()

        content_left = len(content)
        while content_left > 0:
            buffer_left_size = self._chunk_size - self._write_buffer.tell()
            content_part = content[:buffer_left_size]
            content = content[buffer_left_size:]
            self._write_buffer.write(content_part)
            content_left -= len(content_part)
            if self._write_buffer.tell() >= self._chunk_size:
                # write buffer full
                self._flush_to_volume_file()

    def close(self):
        if self._writer is not None:
            if self._write_buffer.tell() > 0:
                self._flush_to_volume_file()
                self._index = 0
            if len(self._meta) > 1:
                meta_content = "\n".join(
                    "{},{}".format(filename, size)
                    for filename, size in self._meta.items()
                )
                # write meta file if file length > 1
                self._writer.write(META_FILE_NAME, to_binary(meta_content))
            self._writer.__exit__(None, None, None)
        if self._reader is not None:
            if self._read_filename is not None:
                # remove read file name
                os.remove(self._read_filename)
            self._reader.__exit__(None, None, None)

    def __enter__(self):
        if self._writer or self._reader:
            (self._writer or self._reader).__enter__()
        return self

    def __exit__(self, *_):
        self.close()


class VolumeFileSystem(FileSystem):
    """
    Schema follows:

    /**project**/volumes/**volume**
    """

    def __init__(self, odps=None, path=None, temp_dir=None, chunk_size=None):
        if odps is None:
            # try to get from environments
            odps = ODPS.from_environments()
        if odps is None:
            # try to get from global config
            odps = ODPS.from_global()
        if odps is None:
            # still None, raise error
            raise ValueError("`odps` should be specified for VolumeFileSystem")

        self._odps = odps
        self._path = path
        self._temp_dir = temp_dir
        self._chunk_size = chunk_size or DEFAULT_CHUNK_SIZE

    @staticmethod
    def parse_from_path(uri):
        parsed = urlparse(uri)
        options = dict()
        if parsed.path:
            options["path"] = parsed.path.strip("/")
        return options

    @staticmethod
    def _extract_info(path, is_file=False):
        options = VolumeFileSystem.parse_from_path(path)
        error_msg = DIR_ERROR_MSG if not is_file else FILE_ERROR_MSG

        if "path" not in options:
            raise ValueError(error_msg.format(path))

        path = options["path"]
        splits = path.split("/")
        expect_size = 3 if not is_file else 4
        if len(splits) != expect_size:
            raise ValueError(error_msg.format(path))

        volumes = splits[1]
        if volumes.lower() != "volumes":
            raise ValueError(error_msg.format(path))

        return splits[:1] + splits[2:]

    def ls(self, path):
        path = path.strip("/") or self._path
        project, volume = self._extract_info(path)
        try:
            files = self._odps.list_volume_partitions(volume, project=project)
        except NoSuchObject:
            raise FileNotFoundError("File {} not found".format(path))
        return [
            "odps:///{}/volumes/{}/{}".format(project, volume, f.name) for f in files
        ]

    def delete(self, path, recursive: bool = False):
        path = path.strip("/") or self._path
        if "/" not in path:
            path = "{}/{}".format(self._path, path)
        options = VolumeFileSystem.parse_from_path(path)
        fpath = options["path"]

        splits = fpath.split("/")
        if len(splits) == 3:
            is_file = False
        elif len(splits) == 4:
            is_file = True
        else:
            raise ValueError(
                "Argument `path` illegal, expect to be "
                "odps:///**project**/volumes/**volume**/ or"
                "odps:///**project**/volumes/**volume**/**partition**/, "
                "got {}".format(path)
            )

        if is_file:
            project, volume, partition = self._extract_info(path, is_file=True)
            if self._odps.exist_volume_partition(
                volume, partition=partition, project=project
            ):
                self._odps.delete_volume_partition(
                    volume, partition=partition, project=project
                )
        else:
            project, volume = self._extract_info(path)
            if self._odps.exist_volume(volume, project=project):
                self._odps.delete_volume(volume, project=project)

    def open(self, path, mode="rb"):
        if mode not in ("rb", "wb"):
            raise ValueError("`mode` can be `rb` or `wb` only")

        path = path.strip("/") or self._path
        if "/" not in path:
            path = "{}/{}".format(self._path, path)
        project, volume, partition = self._extract_info(path, is_file=True)
        volume_parted = self._odps.get_volume_partition(
            volume, partition=partition, project=project
        )

        return _VolumeFileObject(
            volume_parted, mode, temp_dir=self._temp_dir, chunk_size=self._chunk_size
        )

    def mkdir(self, path, create_parents=True):
        assert create_parents

        path = path.strip("/") or self._path
        project, volume = self._extract_info(path)

        if not self._odps.exist_volume(volume, project=project):
            self._odps.create_parted_volume(volume, project=project)

    def stat(self, path):
        path = path.strip("/") or self._path
        if "/" not in path:
            path = "{}/{}".format(self._path, path)
        options = VolumeFileSystem.parse_from_path(path)
        fpath = options["path"]

        splits = fpath.split("/")
        if len(splits) == 3:
            is_file = False
        elif len(splits) == 4:
            is_file = True
        else:
            raise ValueError(
                "Argument `path` illegal, expect to be "
                "odps:///**project**/volumes/**volume**/ or"
                "odps:///**project**/volumes/**volume**/**partition**/, "
                "got {}".format(path)
            )

        if is_file:
            project, volume, partition = self._extract_info(path, is_file=True)
            volume_parted = self._odps.get_volume_partition(
                volume, partition=partition, project=project
            )
            stat = dict(
                name=path,
                size=volume_parted.length,
                created=volume_parted.creation_time,
            )
        else:
            project, volume = self._extract_info(path)
            vol = self._odps.get_volume(volume, project=project)
            stat = dict(name=path, size=vol.length, created=vol.creation_time)
        return stat

    def _isfilestore(self):
        return True

    def cat(self, path):
        raise NotImplementedError

    def exists(self, path):
        raise NotImplementedError

    def glob(self, path, recursive=False):
        raise NotImplementedError

    def isdir(self, path):
        raise NotImplementedError

    def isfile(self, path):
        raise NotImplementedError

    def rename(self, path, new_path):
        raise NotImplementedError

    def walk(self, path):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import XMLRemoteModel
from .. import serializers, utils
from .tasks import Task


class Job(XMLRemoteModel):
    # define __slots__ to keep the property sequence when serializing into a xml
    __slots__ = 'name', 'comment', 'priority', 'running_cluster', 'tasks', 'run_mode'

    _root = 'Job'

    name = serializers.XMLNodeField('Name')
    comment = serializers.XMLNodeField("Comment")
    owner = serializers.XMLNodeField('Owner')
    creation_time = serializers.XMLNodeField('CreationTime',
                                             parse_callback=utils.parse_rfc822)
    last_modified_time = serializers.XMLNodeField('LastModifiedTime',
                                                  parse_callback=utils.parse_rfc822)
    priority = serializers.XMLNodeField('Priority',
                                        parse_callback=int, serialize_callback=int, default=9)
    running_cluster = serializers.XMLNodeField('RunningCluster')
    run_mode = serializers.XMLNodeField('DAG', 'RunMode', default='Sequence')
    tasks = serializers.XMLNodesReferencesField(Task, 'Tasks', '*')

    def add_task(self, task):
        if self.tasks is None:
            self.tasks = []
        self.tasks.append(task)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core import Backend
from ...expr.core import ExprProxy
from ...expr.expressions import Column, CollectionExpr, FilterCollectionExpr
from ...expr.collections import SortedCollectionExpr, SliceCollectionExpr, \
    RowAppliedCollectionExpr, LateralViewCollectionExpr
from ...expr.merge import JoinCollectionExpr, UnionCollectionExpr
from ....models import TableSchema
from ...utils import is_project_expr, traverse_until_source


class ColumnPruning(Backend):
    def __init__(self, dag):
        self._dag = dag

        # As some collections like join and union expr do not have fields,
        # we store them here to illustrate the fields which should keep.
        # Specially, for the JoinCollectionExpr,
        # we store a 2-tuple.
        self._remain_columns = dict()

    def prune(self):
        for node in traverse_until_source(self._dag, top_down=True):
            try:
                node.accept(self)
            except NotImplementedError:
                continue

        return self._dag.root

    def _need_prune(self, expr):
        parents = self._dag.successors(expr)
        if len(parents) == 0:
            return False, set()

        columns = set()
        for p in parents:
            if isinstance(p, Column) and p.input is expr:
                columns.add(p.source_name)
            elif isinstance(p, JoinCollectionExpr):
                proxy = ExprProxy(p)
                if proxy not in self._remain_columns:
                    continue
                lcols, rcols = self._remain_columns[proxy]
                if p._get_child(p._lhs) is expr:
                    columns = columns.union(lcols)
                elif p._get_child(p._rhs) is expr:
                    columns = columns.union(rcols)
                else:
                    continue
            elif isinstance(p, UnionCollectionExpr):
                proxy = ExprProxy(p)
                if proxy not in self._remain_columns:
                    continue
                cols = self._remain_columns[proxy]
                if p._lhs is expr or p._rhs is expr:
                    columns = columns.union(cols)
            elif isinstance(p, LateralViewCollectionExpr):
                proxy = ExprProxy(p)
                if proxy not in self._remain_columns:
                    continue
                col_lists = self._remain_columns[proxy]
                sources = [p.input]
                sources.extend(p.lateral_views)

                found = False
                for clist, coll in zip(col_lists, sources):
                    if coll is expr:
                        columns = columns.union(clist)
                        found = True
                        break
                if not found:
                    continue
            elif hasattr(p, '_collection_resources') and p._collection_resources is not None:
                collection_resources = set([ExprProxy(r) for r in p._collection_resources])
                if ExprProxy(expr) in collection_resources:
                    # the expr is included in apply collection_resources, cannot do any prune
                    return False, set()
            else:
                proxy = ExprProxy(p)
                if proxy not in self._remain_columns:
                    continue
                if hasattr(p, 'input') and p.input is expr:
                    columns = columns.union(self._remain_columns[proxy])

        if len(columns) == 0:
            return False, set()

        if len(set(expr.schema.names) - columns) > 0:
            order = dict((c, i) for i, c in enumerate(expr.schema.names))
            return True, sorted(columns, key=order.get)

        return False, sorted(columns)

    def _need_project_on_source(self, expr):
        def filter_parent(p):
            if not isinstance(p, CollectionExpr):
                return False
            if isinstance(p, RowAppliedCollectionExpr) and p._lateral_view:
                return False
            return True

        def no_project(node):
            if is_project_expr(node):
                # has been projected out, just skip
                return True
            parents = self._dag.successors(node)
            if len(parents) == 1 and isinstance(parents[0], Column):
                return True
            if isinstance(node, (FilterCollectionExpr, SortedCollectionExpr,
                                 SliceCollectionExpr)) or node is expr:
                collection_parents = [n for n in parents if filter_parent(n)]
                if len(collection_parents) == 1 and no_project(collection_parents[0]):
                    return True
            return False

        return not no_project(expr)

    def visit_source_collection(self, expr):
        if not self._need_project_on_source(expr):
            return

        prune, columns = self._need_prune(expr)
        if not prune:
            return

        collection = expr[list(columns)]
        self._dag.substitute(expr, collection)

    def visit_project_collection(self, expr):
        prune, columns = self._need_prune(expr)
        if not prune:
            return

        new_projected = expr.input.select(
            [c for c in expr._fields if c.name in columns])
        self._dag.substitute(expr, new_projected)

    def visit_apply_collection(self, expr):
        # the apply cannot be pruned because the input
        # and output of apply is defined by the user
        return

    def visit_lateral_view(self, expr):
        prune, columns = self._need_prune(expr)
        columns = columns or expr.schema.names
        col_set = set(columns)

        remain_input_cols = [n.source_name for n in expr._fields
                             if isinstance(n, Column) and n.input is expr.input and n.name in col_set]
        input_cols = [remain_input_cols]
        for lv in expr.lateral_views:
            input_cols.append(lv.schema.names)

        self._remain_columns[ExprProxy(expr)] = tuple(input_cols)

        if not prune:
            return
        expr._schema = TableSchema(columns=[expr._schema[col] for col in columns])

    def _visit_all_columns_project_collection(self, expr):
        prune, columns = self._need_prune(expr)
        self._remain_columns[ExprProxy(expr)] = columns or expr.input.schema.names
        if not prune:
            return

        expr._schema = TableSchema(columns=[expr._schema[c] for c in columns])

    def visit_filter_collection(self, expr):
        self._visit_all_columns_project_collection(expr)

    def visit_filter_partition_collection(self, expr):
        prune, columns = self._need_prune(expr)
        if not prune:
            return

        expr._fields = [f for f in expr._fields if f.name in columns]
        expr._schema = TableSchema.from_lists([f.name for f in expr._fields],
                                         [f.dtype for f in expr._fields])

    def visit_slice_collection(self, expr):
        self._visit_all_columns_project_collection(expr)

    def _visit_grouped(self, expr):
        prune, columns = self._need_prune(expr)
        if not prune:
            return

        if expr._fields is None:
            expr._fields = expr._by + expr._aggregations
        expr._fields = [f for f in expr._fields if f.name in columns]
        expr._schema = TableSchema(columns=[expr._schema[c] for c in columns])

    def visit_groupby(self, expr):
        self._visit_grouped(expr)

    def visit_mutate(self, expr):
        self._visit_grouped(expr)

    def visit_reshuffle(self, expr):
        self._visit_all_columns_project_collection(expr)

    def visit_value_counts(self, expr):
        # skip, by and count cannot be pruned
        return

    def visit_sort(self, expr):
        self._visit_all_columns_project_collection(expr)

    def visit_distinct(self, expr):
        # skip, the distinct fields cannot be pruned
        return

    def visit_sample(self, expr):
        if expr._parts is not None:
            self._remain_columns[ExprProxy(expr)] = expr.input.schema.names

    def visit_join(self, expr):
        prune, columns = self._need_prune(expr)

        columns = columns or expr.schema.names
        lhs_cols, rhs_cols = [], []
        for col in columns:
            idx, origin_col = expr._column_origins[col]
            (lhs_cols, rhs_cols)[idx].append(origin_col)

        self._remain_columns[ExprProxy(expr)] = (lhs_cols, rhs_cols)

        if not prune:
            return
        expr._schema = TableSchema(columns=[expr._schema[col] for col in columns])

    def visit_union(self, expr):
        prune, columns = self._need_prune(expr)
        self._remain_columns[ExprProxy(expr)] = columns or expr.schema.names
        if not prune:
            return

        expr._schema = TableSchema(columns=[expr._schema[c] for c in columns])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zlib

from .compat import six
from .config import options


class Crc32(object):
    def __init__(self):
        self.crc = None

    def update(self, buf, off=None, length=None):
        assert isinstance(buf, (six.binary_type, bytearray))
        off = off or 0
        length = length or len(buf)
        to_crc = buf[off: off+length]
        if isinstance(to_crc, bytearray):
            to_crc = bytes(to_crc)
        if self.crc:
            self.crc = zlib.crc32(to_crc, self.crc)
        else:
            self.crc = zlib.crc32(to_crc)

    def reset(self):
        self.crc = None

    def getvalue(self):
        return self.crc


try:
    if not options.force_py:
        from .src.crc32c_c import Crc32c
    else:
        Crc32c = None
except ImportError as e:
    if options.force_c:
        raise e
    Crc32c = None


if Crc32c is None:
    _CRC_TABLE = (
        0x00000000, 0xf26b8303, 0xe13b70f7, 0x1350f3f4,
        0xc79a971f, 0x35f1141c, 0x26a1e7e8, 0xd4ca64eb,
        0x8ad958cf, 0x78b2dbcc, 0x6be22838, 0x9989ab3b,
        0x4d43cfd0, 0xbf284cd3, 0xac78bf27, 0x5e133c24,
        0x105ec76f, 0xe235446c, 0xf165b798, 0x030e349b,
        0xd7c45070, 0x25afd373, 0x36ff2087, 0xc494a384,
        0x9a879fa0, 0x68ec1ca3, 0x7bbcef57, 0x89d76c54,
        0x5d1d08bf, 0xaf768bbc, 0xbc267848, 0x4e4dfb4b,
        0x20bd8ede, 0xd2d60ddd, 0xc186fe29, 0x33ed7d2a,
        0xe72719c1, 0x154c9ac2, 0x061c6936, 0xf477ea35,
        0xaa64d611, 0x580f5512, 0x4b5fa6e6, 0xb93425e5,
        0x6dfe410e, 0x9f95c20d, 0x8cc531f9, 0x7eaeb2fa,
        0x30e349b1, 0xc288cab2, 0xd1d83946, 0x23b3ba45,
        0xf779deae, 0x05125dad, 0x1642ae59, 0xe4292d5a,
        0xba3a117e, 0x4851927d, 0x5b016189, 0xa96ae28a,
        0x7da08661, 0x8fcb0562, 0x9c9bf696, 0x6ef07595,
        0x417b1dbc, 0xb3109ebf, 0xa0406d4b, 0x522bee48,
        0x86e18aa3, 0x748a09a0, 0x67dafa54, 0x95b17957,
        0xcba24573, 0x39c9c670, 0x2a993584, 0xd8f2b687,
        0x0c38d26c, 0xfe53516f, 0xed03a29b, 0x1f682198,
        0x5125dad3, 0xa34e59d0, 0xb01eaa24, 0x42752927,
        0x96bf4dcc, 0x64d4cecf, 0x77843d3b, 0x85efbe38,
        0xdbfc821c, 0x2997011f, 0x3ac7f2eb, 0xc8ac71e8,
        0x1c661503, 0xee0d9600, 0xfd5d65f4, 0x0f36e6f7,
        0x61c69362, 0x93ad1061, 0x80fde395, 0x72966096,
        0xa65c047d, 0x5437877e, 0x4767748a, 0xb50cf789,
        0xeb1fcbad, 0x197448ae, 0x0a24bb5a, 0xf84f3859,
        0x2c855cb2, 0xdeeedfb1, 0xcdbe2c45, 0x3fd5af46,
        0x7198540d, 0x83f3d70e, 0x90a324fa, 0x62c8a7f9,
        0xb602c312, 0x44694011, 0x5739b3e5, 0xa55230e6,
        0xfb410cc2, 0x092a8fc1, 0x1a7a7c35, 0xe811ff36,
        0x3cdb9bdd, 0xceb018de, 0xdde0eb2a, 0x2f8b6829,
        0x82f63b78, 0x709db87b, 0x63cd4b8f, 0x91a6c88c,
        0x456cac67, 0xb7072f64, 0xa457dc90, 0x563c5f93,
        0x082f63b7, 0xfa44e0b4, 0xe9141340, 0x1b7f9043,
        0xcfb5f4a8, 0x3dde77ab, 0x2e8e845f, 0xdce5075c,
        0x92a8fc17, 0x60c37f14, 0x73938ce0, 0x81f80fe3,
        0x55326b08, 0xa759e80b, 0xb4091bff, 0x466298fc,
        0x1871a4d8, 0xea1a27db, 0xf94ad42f, 0x0b21572c,
        0xdfeb33c7, 0x2d80b0c4, 0x3ed04330, 0xccbbc033,
        0xa24bb5a6, 0x502036a5, 0x4370c551, 0xb11b4652,
        0x65d122b9, 0x97baa1ba, 0x84ea524e, 0x7681d14d,
        0x2892ed69, 0xdaf96e6a, 0xc9a99d9e, 0x3bc21e9d,
        0xef087a76, 0x1d63f975, 0x0e330a81, 0xfc588982,
        0xb21572c9, 0x407ef1ca, 0x532e023e, 0xa145813d,
        0x758fe5d6, 0x87e466d5, 0x94b49521, 0x66df1622,
        0x38cc2a06, 0xcaa7a905, 0xd9f75af1, 0x2b9cd9f2,
        0xff56bd19, 0x0d3d3e1a, 0x1e6dcdee, 0xec064eed,
        0xc38d26c4, 0x31e6a5c7, 0x22b65633, 0xd0ddd530,
        0x0417b1db, 0xf67c32d8, 0xe52cc12c, 0x1747422f,
        0x49547e0b, 0xbb3ffd08, 0xa86f0efc, 0x5a048dff,
        0x8ecee914, 0x7ca56a17, 0x6ff599e3, 0x9d9e1ae0,
        0xd3d3e1ab, 0x21b862a8, 0x32e8915c, 0xc083125f,
        0x144976b4, 0xe622f5b7, 0xf5720643, 0x07198540,
        0x590ab964, 0xab613a67, 0xb831c993, 0x4a5a4a90,
        0x9e902e7b, 0x6cfbad78, 0x7fab5e8c, 0x8dc0dd8f,
        0xe330a81a, 0x115b2b19, 0x020bd8ed, 0xf0605bee,
        0x24aa3f05, 0xd6c1bc06, 0xc5914ff2, 0x37faccf1,
        0x69e9f0d5, 0x9b8273d6, 0x88d28022, 0x7ab90321,
        0xae7367ca, 0x5c18e4c9, 0x4f48173d, 0xbd23943e,
        0xf36e6f75, 0x0105ec76, 0x12551f82, 0xe03e9c81,
        0x34f4f86a, 0xc69f7b69, 0xd5cf889d, 0x27a40b9e,
        0x79b737ba, 0x8bdcb4b9, 0x988c474d, 0x6ae7c44e,
        0xbe2da0a5, 0x4c4623a6, 0x5f16d052, 0xad7d5351,
    )

    _CRC_INIT = 0xffffffff


    class Crc32c(object):
        _method = 'py'

        def __init__(self):
            self.crc = _CRC_INIT

        def update(self, buf, off=None, length=None):
            """
            :param buf: buf to update
            :type buf: bytearray
            :param off: offset
            :param length: length
            """

            off = off or 0
            length = length or len(buf)

            to_crc = buf[off: off + length]

            crc = self.crc
            for b in to_crc:
                table_index = (crc ^ b) & 0xff
                crc = (_CRC_TABLE[table_index] ^ (crc >> 8)) & 0xffffffff
            self.crc = crc & 0xffffffff

        def reset(self):
            self.crc = _CRC_INIT

        @classmethod
        def _crc_finalize(cls, crc):
            return crc ^ 0xffffffff

        def getvalue(self):
            return Crc32c._crc_finalize(self.crc)
